from typing import List, Literal, Optional

# We don't use it with untrusted random input
from xml.etree.ElementTree import Element, SubElement, tostring  # nosec: B405

from httpx import Client as HTTPXClient, Response, Timeout
from xmltodict import parse as parse_xml  # type: ignore[import]

from smartpost.errors import ShipmentOrderError
from smartpost.models import Destination, OrderInfo, ShipmentOrder


class Client:
    """Synchronous SmartPost API client that takes care of all low-level things."""

    def __init__(
        self,
        username: str = "",
        password: str = "",  # nosec: B107
        *,
        read_timeout: int = 10,
    ) -> None:
        self._read_timeout = read_timeout
        self._client: Optional[HTTPXClient] = None

        # XML element "authentication" will be sent with requests that require auth

        self._auth = Element("authentication")
        user_el = SubElement(self._auth, "user")
        password_el = SubElement(self._auth, "password")
        user_el.text = username
        password_el.text = password

    @property
    def client(self) -> HTTPXClient:
        if not self._client:
            self._client = HTTPXClient(
                base_url="https://iseteenindus.smartpost.ee/api",
                http2=True,
                timeout=Timeout(5, read=self._read_timeout),
            )

        return self._client

    def get(self, request: str, **kwargs: str) -> Response:
        return self.client.get("/", params={"request": request, **kwargs})

    def post(self, request: str, xml_content: bytes) -> Response:
        return self.client.post("/", params={"request": request}, content=xml_content)

    def get_ee_terminals(self) -> List[Destination]:
        """Fetches list of all Estonia terminals.

        Returns:
            A list of `Destination` instances representing all Estonia terminals.
        """
        response = self.get("destinations", country="EE", type="APT")
        # TODO: Add request errors handling
        destinations = parse_xml(response.read(), force_list=("item",))
        return [Destination(**item) for item in destinations["destinations"]["item"]]

    def get_ee_express_terminals(self) -> List[Destination]:
        """Fetches list of all Estonia express terminals.

        Returns:
            A list of `Destination` instances
            representing all Estonia express terminals.
        """
        response = self.get("destinations", country="EE", type="APT", filter="express")
        # TODO: Add request errors handling
        destinations = parse_xml(response.read(), force_list=("item",))
        return [Destination(**item) for item in destinations["destinations"]["item"]]

    def get_fi_terminals(self) -> List[Destination]:
        """Fetches list of all Finland terminals.

        Returns:
            A list of `Destination` instances representing all Finland terminals.
        """
        response = self.get("destinations", country="FI", type="APT")
        # TODO: Add request errors handling
        destinations = parse_xml(response.read(), force_list=("item",))
        return [Destination(**item) for item in destinations["destinations"]["item"]]

    def get_fi_post_offices(self) -> List[Destination]:
        """Fetches list of all Finland post offices.

        Returns:
            A list of `Destination` instances representing all Finland post offices.
        """
        response = self.get("destinations", country="FI", type="PO")
        # TODO: Add request errors handling
        destinations = parse_xml(response.read(), force_list=("item",))
        return [Destination(**item) for item in destinations["destinations"]["item"]]

    def add_shipment_orders(
        self,
        shipment_orders: List[ShipmentOrder],
        report_emails: Optional[List[str]] = None,
    ) -> List[OrderInfo]:
        """Adds shipment orders to SmartPost system.

        Args:
            shipment_orders:
                a list of `ShipmentOrder` instances representing orders to be added.
            report_emails:
                optional list of strings with emails to which
                reports about order will be sent.

        Returns:
            A list of `OrderInfo` instances representing all added orders.

        Raises:
            ShipmentOrderError:
                SmartPost API had issues with shipment orders you sent.
        """
        document = Element("orders")
        document.insert(0, self._auth)
        report_el = SubElement(document, "report")
        for email in report_emails or []:
            email_el = SubElement(report_el, "email")
            email_el.text = email

        document.extend(order.to_xml() for order in shipment_orders)
        response = self.post("shipment", tostring(document))
        if response.status_code == 400:
            errors = parse_xml(response.read(), force_list=("item",))
            raise ShipmentOrderError(errors)

        orders = parse_xml(response.read(), force_list=("item",))
        return [OrderInfo(**order) for order in orders["orders"]["item"]]

    def get_labels_pdf(
        self,
        format: Literal["A5", "A6", "A6-4", "A7", "A7-8"],
        barcodes: List[str],
    ) -> bytes:
        """Requests PDF file with labels for orders.

        Args:
            format:
                a string with PDF page format, any of the following:
                A5, A6, A6-4, A7, A7-8.
            barcodes:
                a list of strings with order barcodes for which labels are requested.

        Returns:
            A PDF file bytes. File will contain a page for each barcode provided in
            `barcodes` argument, all pages will be in format from `format` argument.

        Raises:
            httpx.ReadTimeout:
                SmartPost API did not manage to send PDF file in time.
        """
        document = Element("labels")
        document.insert(0, self._auth)
        format_el = SubElement(document, "format")
        format_el.text = format
        for barcode in barcodes:
            barcode_el = SubElement(document, "barcode")
            barcode_el.text = barcode

        response = self.post("labels", tostring(document))
        # TODO: Add request errors handling
        return response.read()


from typing import Any, Dict, List, Optional

errors_explanation: Dict[str, str] = {
    "000": "Destination info missing",
    "001": "Required input missing",
    "002": "Barcode already exists",
    "003": "Not a phone number",
    "004": "Not an e-mail address",
    "005": "Not numeric",
    "006": "Unknown destination (place_id)",
    "007": "Express checked but destination is not a express APT",
    "008": "Multiple destinations set",
    "009": "Courier service timeframe missing",
    "010": "Courier service city missing",
    "011": "Courier service postal code missing",
    "012": "Courier address missing",
    "013": "COD used but service not available",
    "014": "Invalid door size",
    "015": "String too long (LQ items)",
    "016": "Integer out of bounds (LQ items)",
    "017": "Value not in enum list (LQ items)",
    "018": "Sending with door code service not activated",
}


class ShipmentOrderErrorDetails:
    """Additional details about shipment order with issues."""

    def __init__(self, item: Dict[str, Any]) -> None:
        error: Dict[str, Any] = item["error"]
        self.barcode: Optional[str] = item["barcode"]
        self.reference: Optional[str] = item["reference"]
        self.code: str = error["code"]
        self.message: str = errors_explanation[error["code"]]
        self.text: str = error["text"]
        self.input: Optional[str] = error["input"]
        # Field "field" is not mentioned in docs (optional)
        self.field: Optional[str] = error.get("field")

    def __str__(self) -> str:
        return (
            f"ShipmentOrderErrorDetails(barcode={self.barcode}, "
            f"reference={self.reference}, code={self.code}, message={self.message}, "
            f"text={self.text}, input={self.input}, field={self.field})"
        )

    __repr__ = __str__


class ShipmentOrderError(Exception):
    """Error that is raised when there are issues with shipment orders."""

    def __init__(self, errors: Dict[str, Any]) -> None:
        self.errors: List[ShipmentOrderErrorDetails] = [
            ShipmentOrderErrorDetails(item) for item in errors["orders"]["item"]
        ]



import os
import re

from setuptools import find_packages, setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


def get_long_description():
    """Return the README"""
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="aiosmartpost",
    python_requires=">=3.8",
    version=get_version("smartpost"),
    description="Itella SmartPost API wrapper for humans 📦",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/HarrySky/aiosmartpost",
    license="Unlicense",
    author="Igor Nehoroshev",
    author_email="mail@neigor.me",
    packages=find_packages(exclude=["tests"]),
    # Use MANIFEST.in for data files
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        # For HTTP requests to API using HTTP/2
        "httpx",
        "h2",
        # For parsing API responses to dict
        # (SmartPost API uses XML instead of JSON for some reason)
        "xmltodict",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: The Unlicense (Unlicense)",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
)



from dataclasses import dataclass, field
from typing import Literal, Optional, TypedDict, Union

# We don't use it with untrusted random input
from xml.etree.ElementTree import Element, SubElement  # nosec: B405


@dataclass
class Destination:
    place_id: int
    name: str
    city: str
    address: str
    country: Literal["EE", "FI"]
    postalcode: str
    routingcode: str
    availability: str
    description: str
    lat: float
    lng: float

    def __post_init__(self) -> None:
        # Values passed to __init__ are expected to be str (XML)
        self.place_id = int(self.place_id)
        self.lat = float(self.lat)
        self.lng = float(self.lng)


@dataclass
class Sender:
    name: str
    phone: str
    email: str
    #: Amount of money the sender has to pay before sending the parcel
    cash: float
    #: Business customer’s account identificator
    account: int


@dataclass
class Recipient:
    name: str
    phone: str
    email: str
    #: Amount of money the recipient has to pay before receiving the parcel
    cash: Optional[float] = None
    #: ID code of the recipient; needed if ID validation is required.
    idcode: Optional[int] = None

    def to_xml(self) -> Element:
        recipient = Element("recipient")
        name = SubElement(recipient, "name")
        name.text = self.name
        phone = SubElement(recipient, "phone")
        phone.text = self.phone
        email = SubElement(recipient, "email")
        email.text = self.email

        # Optional fields handling

        if self.cash is not None:
            cash = SubElement(recipient, "cash")
            cash.text = str(self.cash)

        if self.idcode is not None:
            idcode = SubElement(recipient, "idcode")
            idcode.text = str(self.idcode)

        return recipient


@dataclass
class EETerminalDestination:
    place_id: int

    def to_xml(self) -> Element:
        destination = Element("destination")
        place_id = SubElement(destination, "place_id")
        place_id.text = str(self.place_id)
        return destination


@dataclass
class EECourierDestination:
    street: str
    house: str
    apartment: str
    city: str
    country: str
    postalcode: str
    details: str
    timewindow: Literal[1, 2, 3]

    def to_xml(self) -> Element:
        # TODO: IMPLEMENT
        raise NotImplementedError()


@dataclass
class FIDestination:
    postalcode: str
    routingcode: str

    def to_xml(self) -> Element:
        destination = Element("destination")
        postalcode = SubElement(destination, "postalcode")
        postalcode.text = self.postalcode
        routingcode = SubElement(destination, "routingcode")
        routingcode.text = self.routingcode
        return destination


ShipmentDestination = Union[EETerminalDestination, EECourierDestination, FIDestination]


@dataclass
class ShipmentOrder:
    recipient: Recipient
    destination: ShipmentDestination
    #: Barcode will be generated if not specified
    barcode: Optional[str] = None
    reference: Optional[str] = None
    content: Optional[str] = None
    orderparent: Optional[str] = None
    weight: Optional[float] = None
    size: Optional[Literal[5, 6, 7, 8, 11]] = None
    #: Only needed when "sending with door code" service is used
    sender: Optional[Sender] = None
    # TODO: Create proper types
    lqitems: None = None
    additionalservices: None = None

    def to_xml(self) -> Element:
        item = Element("item")
        recipient_el = self.recipient.to_xml()
        item.append(recipient_el)
        destination_el = self.destination.to_xml()
        item.append(destination_el)

        # Optional fields handling

        if self.barcode:
            barcode = SubElement(item, "barcode")
            barcode.text = self.barcode

        if self.reference:
            reference = SubElement(item, "reference")
            reference.text = self.reference

        if self.content:
            content = SubElement(item, "content")
            content.text = self.content

        if self.orderparent:
            orderparent = SubElement(item, "orderparent")
            orderparent.text = self.orderparent

        if self.weight:
            weight = SubElement(item, "weight")
            weight.text = str(self.weight)

        if self.size:
            size = SubElement(item, "size")
            size.text = str(self.size)

        # TODO: Deal with sender, lqitems and additionalservices

        return item


class SenderDoorCode(TypedDict):
    doorcode: int


@dataclass
class OrderInfo:
    barcode: str
    reference: str
    sender: Optional[SenderDoorCode] = None

    #: Convinient access to `sender["doorcode"]`
    doorcode: Optional[int] = field(default=None, init=False)

    def __post_init__(self) -> None:
        if self.sender:
            # Values passed to __init__ are expected to be str (XML)
            self.doorcode = int(self.sender["doorcode"])
            self.sender["doorcode"] = self.doorcode
